import os
from unittest.mock import MagicMock
import requests

from torch.utils.data import IterableDataset


def train_tokenizer(destination_path):
    destination_path.mkdir(parents=True, exist_ok=True)

    # download the tiny shakespeare dataset
    input_file_path = destination_path / "input.txt"
    if not input_file_path.exists():
        data_url = "https://raw.githubusercontent.com/karpathy/char-rnn/master/data/tinyshakespeare/input.txt"
        with open(input_file_path, "w") as f:
            f.write(requests.get(data_url).text)

    from lit_llama import Tokenizer
    Tokenizer.train(
        input=input_file_path,
        destination=destination_path,
        vocab_size=100,
    )

    return destination_path / "tokenizer.model"


def test_packed_dataset(tmp_path):
    tokenizer_path = train_tokenizer(tmp_path)

    from lit_llama import Tokenizer
    tokenizer = Tokenizer(tokenizer_path)

    texts = [
      "The moment of truth is upon us.",
      "Time to open the fridge."
    ]

    from lit_llama.packed_dataset import PackedDatasetBuilder, PackedDataset, HDR_SIZE

    block_size = 10
    n_blocks = 2
    chunk_size = block_size * n_blocks

    builder = PackedDatasetBuilder(
        outdir=tmp_path,
        prefix="packed_dataset",
        chunk_size=chunk_size,
        sep_token=tokenizer.bos_id,
        dtype="auto",
        vocab_size=100,
    )

    text_ids = []

    for text in texts:
        text_ids = tokenizer.encode(text)
        assert text_ids[0] == tokenizer.bos_id
        builder.add_array(text_ids)

    filenames = builder.filenames

    assert len(filenames) == 2
    assert os.path.basename(filenames[0]) == "packed_dataset_0000000000.bin"
    assert os.path.basename(filenames[1]) == "packed_dataset_0000000001.bin"

    import numpy as np

    ex_tokenized = [
        tokenizer.encode(text).numpy().astype(builder.dtype)
        for text in texts
    ]
    ex_tokenized = np.concatenate(ex_tokenized)
    ex_tokenized = ex_tokenized[:2 * chunk_size]

    for filename, el in zip(filenames, np.array_split(ex_tokenized, 2)):
        mmap = np.memmap(filename, mode="r", order="C", offset=HDR_SIZE)
        count = len(mmap) // np.dtype(builder.dtype).itemsize
        arr = np.frombuffer(
            mmap, dtype=builder.dtype, count=count, offset=0
        )
        where_bos = np.where(arr == tokenizer.bos_id)
        # we expect two BOS tokens, one per file
        assert len(where_bos) == 1
        assert np.array_equal(arr, el)

    dataset = PackedDataset(filenames=filenames, n_chunks=2, block_size=block_size, shuffle=False)

    ex_split = np.array_split(ex_tokenized, ex_tokenized.shape[0] // block_size)

    for item, el in zip(dataset, ex_split):
        assert np.array_equal(item, el)

    dataset = PackedDataset(filenames=filenames, n_chunks=2, block_size=block_size, seed=12345)

    for i, item in enumerate(dataset):
        block_idxs = iter(dataset)._block_idxs
        assert np.array_equal(item, ex_split[block_idxs[i]])

    dataset = PackedDataset(filenames=filenames, n_chunks=2, block_size=block_size, seed=12345, wrap=True)

    for i, item in enumerate(dataset):
        if i > 24:
            break

    dataset = PackedDataset(filenames=filenames, n_chunks=1, block_size=block_size, seed=12345)

    for i, item in enumerate(dataset):
        block_idxs = iter(dataset)._block_idxs
        chunk_idx = i // n_blocks * n_blocks
        assert np.array_equal(item, ex_split[chunk_idx + block_idxs[i % n_blocks]])

    block_size_ = block_size // 2
    ex_split = np.array_split(ex_tokenized, ex_tokenized.shape[0] // block_size_)
    dataset = PackedDataset(filenames=filenames, n_chunks=2, block_size=block_size_, seed=12345)

    for i, item in enumerate(dataset):
        block_idxs = iter(dataset)._block_idxs
        assert np.array_equal(item, ex_split[block_idxs[i]])

    block_size_ = block_size // 3
    n_chunks = 2
    ex_chunks = np.split(ex_tokenized, n_chunks)
    n_splits = ex_tokenized.shape[0] // n_chunks // block_size_
    ex_splits = [np.split(el[:n_splits * block_size_], n_splits) for el in ex_chunks]
    ex_split = sum(ex_splits, [])

    dataset = PackedDataset(filenames=filenames, n_chunks=n_chunks, block_size=block_size_, seed=12345)

    for i, item in enumerate(dataset):
        block_idxs = iter(dataset)._block_idxs
        assert np.array_equal(item, ex_split[block_idxs[i]])


class SimpleDataset(IterableDataset):
    def __init__(self, start, end):
        super().__init__()
        self._start = start
        self._end = end

    def __iter__(self):
        return iter(range(self._start, self._end))
        

def test_combined_dataset(tmp_path):
    from lit_llama.packed_dataset import CombinedDataset

    dataset1 = SimpleDataset(0, 10)
    dataset2 = SimpleDataset(10, 20)
    dataset = CombinedDataset(datasets=[dataset1, dataset2], weights=[1.0, 0.0], seed=12345)

    res = [el for el in dataset]
    assert res == list(range(0, 10))

    dataset1 = SimpleDataset(0, 10)
    dataset2 = SimpleDataset(10, 20)
    dataset = CombinedDataset(datasets=[dataset1, dataset2], weights=[0.0, 1.0], seed=12345)

    res = [el for el in dataset]
    assert res == list(range(10, 20))

    dataset1 = SimpleDataset(0, 10)
    dataset2 = SimpleDataset(10, 20)
    dataset = CombinedDataset(datasets=[dataset1, dataset2], weights=[0.5, 0.5], seed=12345)

    res = [el for el in dataset]
    assert 9 in res or 19 in res
    if len(res) > 10:
        assert 0 in res and 10 in res


def test_sharded_packed_dataset(monkeypatch):
    import lit_llama.packed_dataset
    from lit_llama.packed_dataset import PackedDataset

    dataset_iterator_mock = MagicMock()
    monkeypatch.setattr(lit_llama.packed_dataset, "PackedDatasetIterator", dataset_iterator_mock)
    filenames = [str(i) for i in range(10)]

    # world_size = 1, rank = 0
    iter(PackedDataset(filenames=filenames, n_chunks=2, block_size=2))
    assert dataset_iterator_mock.call_args[1]["filenames"] == filenames
    dataset_iterator_mock.reset_mock()
    # world_size = 2, rank = 0
    iter(PackedDataset(filenames=filenames, n_chunks=2, block_size=2, num_processes=2, process_rank=0))
    assert dataset_iterator_mock.call_args[1]["filenames"] == ["0", "2", "4", "6", "8"]
    dataset_iterator_mock.reset_mock()
    # world_size = 2, rank = 1
    iter(PackedDataset(filenames=filenames, n_chunks=2, block_size=2, num_processes=2, process_rank=1))
    assert dataset_iterator_mock.call_args[1]["filenames"] == ["1", "3", "5", "7", "9"]
    dataset_iterator_mock.reset_mock()
    
    # world_size = 3, rank = 0 (dataset size not cleanly divisible by world size)
    iter(PackedDataset(filenames=filenames, n_chunks=2, block_size=2, num_processes=3, process_rank=0))
    assert dataset_iterator_mock.call_args[1]["filenames"] == ["0", "3", "6"]
    dataset_iterator_mock.reset_mock()
    # world_size = 3, rank = 1 (dataset size not cleanly divisible by world size)
    iter(PackedDataset(filenames=filenames, n_chunks=2, block_size=2, num_processes=3, process_rank=1))
    assert dataset_iterator_mock.call_args[1]["filenames"] == ["1", "4", "7"]
    dataset_iterator_mock.reset_mock()
    # world_size = 3, rank = 2 (dataset size not cleanly divisible by world size)
    iter(PackedDataset(filenames=filenames, n_chunks=2, block_size=2, num_processes=3, process_rank=2))
    assert dataset_iterator_mock.call_args[1]["filenames"] == ["2", "5", "8"]



import os

from setuptools import setup, find_packages


_PATH_ROOT = os.path.dirname(__file__)

with open(os.path.join(_PATH_ROOT, "README.md"), encoding="utf-8") as fo:
    readme = fo.read()

setup(
    name='lit-llama',
    version='0.1.0',
    description='Implementation of the LLaMA language model',
    author='Lightning AI',
    url='https://github.com/lightning-AI/lit-llama',
    install_requires=[
        "torch>=2.0.0",
        "lightning @ git+https://github.com/Lightning-AI/lightning@master",
        "sentencepiece",
        "bitsandbytes",
    ],
    packages=find_packages(),
    long_description=readme,
    long_description_content_type="text/markdown",
)



import torch
import pytest
import sys


def copy_mlp(llama_mlp, orig_llama_mlp) -> None:
    orig_llama_mlp.w1.weight.copy_(llama_mlp.c_fc1.weight)
    orig_llama_mlp.w3.weight.copy_(llama_mlp.c_fc2.weight)
    orig_llama_mlp.w2.weight.copy_(llama_mlp.c_proj.weight)


def copy_attention(llama_attn, orig_llama_attn) -> None:
    n_embd = llama_attn.c_attn.weight.shape[1]
    orig_llama_attn.wq.weight.copy_(llama_attn.c_attn.weight[:n_embd])
    orig_llama_attn.wk.weight.copy_(llama_attn.c_attn.weight[n_embd:-n_embd])
    orig_llama_attn.wv.weight.copy_(llama_attn.c_attn.weight[-n_embd:])
    orig_llama_attn.wo.weight.copy_(llama_attn.c_proj.weight)


def copy_block(llama_block, orig_llama_block) -> None:
    orig_llama_block.attention_norm.weight.copy_(llama_block.rms_1.scale)
    copy_attention(llama_block.attn, orig_llama_block.attention)
    orig_llama_block.ffn_norm.weight.copy_(llama_block.rms_2.scale)
    copy_mlp(llama_block.mlp, orig_llama_block.feed_forward)


def copy_weights(llama_model, orig_llama_model) -> None:
    orig_llama_model.tok_embeddings.weight.copy_(llama_model.transformer.wte.weight)
    for llama_block, orig_llama_block in zip(llama_model.transformer.h, orig_llama_model.layers):
        copy_block(llama_block, orig_llama_block)
    orig_llama_model.norm.weight.copy_(llama_model.transformer.ln_f.scale)
    orig_llama_model.output.weight.copy_(llama_model.lm_head.weight)


@torch.no_grad()
@pytest.mark.parametrize("kv_cache", (False, True))
def test_to_orig_llama(lit_llama, orig_llama, kv_cache) -> None:
    block_size = 64
    vocab_size = 32000
    n_layer = 16
    n_head = 16
    n_embd = 32
    batch_size = 3

    llama_config = lit_llama.LLaMAConfig(
        block_size=block_size, vocab_size=vocab_size, n_layer=n_layer, n_head=n_head, n_embd=n_embd
    )
    orig_llama_config = orig_llama.ModelArgs(
        dim=n_embd,
        n_layers=n_layer,
        n_heads=n_head,
        vocab_size=vocab_size,
        norm_eps=1e-5,
        max_seq_len=block_size,
        max_batch_size=batch_size,
    )

    seq_len = orig_llama_config.max_seq_len
    token_sample = torch.randint(0, orig_llama_config.vocab_size, size=(batch_size, seq_len), dtype=torch.int64)

    llama_model = lit_llama.LLaMA(llama_config)
    llama_model.apply(llama_model._init_weights)
    orig_llama_model = orig_llama.Transformer(orig_llama_config)

    copy_weights(llama_model, orig_llama_model)

    orig_llama_embed = orig_llama_model.tok_embeddings(token_sample)
    llama_embed = llama_model.transformer.wte(token_sample)
    assert torch.allclose(orig_llama_embed, llama_embed)

    llama_rope = llama_model.build_rope_cache(token_sample)
    llama_mask = llama_model.build_mask_cache(token_sample)
    orig_llama_mask = torch.full((1, 1, seq_len, seq_len), float("-inf"))
    orig_llama_mask = torch.triu(orig_llama_mask, diagonal=1)
    if kv_cache:
        orig_llama_block_out = orig_llama_model.layers[0](
            orig_llama_embed, 0, orig_llama_model.freqs_cis[:seq_len], orig_llama_mask
        )
        theirs_k_cache = orig_llama_model.layers[0].attention.cache_k
        theirs_v_cache = orig_llama_model.layers[0].attention.cache_v
        head_size = n_embd // n_head
        kv_cache_shape = (batch_size, n_head, block_size, head_size)
        ours_kv_cache = torch.zeros(kv_cache_shape), torch.zeros(kv_cache_shape)
        (llama_block_out, ours_kv_cache) = llama_model.transformer.h[0](
            llama_embed, llama_rope, llama_mask, seq_len, torch.arange(block_size), ours_kv_cache
        )
        ours_k_cache = ours_kv_cache[0].permute(0, 2, 1, 3)
        ours_v_cache = ours_kv_cache[1].permute(0, 2, 1, 3)
        torch.testing.assert_close(ours_k_cache, theirs_k_cache)
        torch.testing.assert_close(ours_v_cache, theirs_v_cache)
    else:
        orig_llama_block_out = orig_llama_model.layers[0](
            orig_llama_embed, 0, orig_llama_model.freqs_cis[:seq_len], orig_llama_mask
        )
        (llama_block_out, _) = llama_model.transformer.h[0](llama_embed, llama_rope, llama_mask, seq_len)
    assert torch.allclose(orig_llama_block_out, llama_block_out)

    expected = orig_llama_model(token_sample, 0)
    out = llama_model(token_sample)
    assert torch.allclose(out, expected)


@pytest.mark.skipif(not torch.cuda.is_available(), reason="Requires CUDA")
@torch.no_grad()
def test_bfloat16_llama_init(lit_llama, orig_llama) -> None:
    from lit_llama.utils import EmptyInitOnDevice

    block_size = 64
    vocab_size = 32000
    n_layer = 16
    n_head = 16
    n_embd = 32

    llama_config = lit_llama.LLaMAConfig(
        block_size=block_size, vocab_size=vocab_size, n_layer=n_layer, n_head=n_head, n_embd=n_embd
    )
    llama_model = lit_llama.LLaMA(llama_config)
    llama_model.apply(llama_model._init_weights)

    batch_size = 3

    token_sample = torch.randint(0, vocab_size, size=(batch_size, block_size), dtype=torch.int64)

    expected = llama_model(token_sample)

    with EmptyInitOnDevice(device="cuda", dtype=torch.bfloat16):
        llama_model2 = lit_llama.LLaMA(llama_config)
    llama_model2.load_state_dict(llama_model.state_dict(keep_vars=True))

    out = llama_model2(token_sample.cuda()).float().cpu()
    torch.testing.assert_close(out, expected, atol=5e-3, rtol=1e-3)


def copy_adapter_weights(llama_model, orig_llama_model) -> None:
    # copy the gating parameter
    for llama_block, orig_llama_block in zip(llama_model.transformer.h, orig_llama_model.layers):
        if hasattr(llama_block.attn, "gating_factor"):
            llama_block.attn.gating_factor.copy_(orig_llama_block.attention.gate)

    # In the original model, there is one embedding layer for all blocks combined
    orig_adapter_wte = orig_llama_model.adapter_query.weight.reshape(
        orig_llama_model.params.adapter_layer, orig_llama_model.params.adapter_len, orig_llama_model.params.dim
    )

    # In ours, the embedding layer is split across the individual attention layers
    index = 0
    for llama_block in llama_model.transformer.h:
        if hasattr(llama_block.attn, "adapter_wte"):
            llama_block.attn.adapter_wte.weight.copy_(orig_adapter_wte[index])
            index += 1


def enable_gate(model):
    for name, param in model.named_parameters():
        if "gating_factor" in name or "gate" in name:
            param.fill_(1)


@torch.no_grad()
def test_adapter_parity(orig_llama_adapter):
    """Test parity between our implementation of LLaMA-Adapter and the reference code."""
    import lit_llama.adapter as lit_llama

    orig_llama = orig_llama_adapter

    block_size = 32
    vocab_size = 100
    n_layer = 2
    n_head = 4
    n_embd = 16
    adapter_prompt_length: int = 10
    adapter_start_layer: int = 0

    llama_config = lit_llama.LLaMAConfig(
        block_size=block_size,
        vocab_size=vocab_size,
        n_layer=n_layer,
        n_head=n_head,
        n_embd=n_embd,
        adapter_prompt_length=adapter_prompt_length,
        adapter_start_layer=adapter_start_layer,
    )
    orig_llama_config = orig_llama.ModelArgs(
        dim=n_embd,
        n_layers=n_layer,
        n_heads=n_head,
        vocab_size=vocab_size,
        norm_eps=1e-5,
        max_seq_len=block_size,
        adapter_len=adapter_prompt_length,
        adapter_layer=(n_layer - adapter_start_layer),
    )

    batch_size = 3
    token_sample = torch.randint(
        0, orig_llama_config.vocab_size, size=(batch_size, orig_llama_config.max_seq_len), dtype=torch.int64
    )

    llama_model = lit_llama.LLaMA(llama_config)
    llama_model.apply(llama_model._init_weights)
    orig_llama_model = orig_llama.Transformer(orig_llama_config)

    copy_weights(llama_model, orig_llama_model)
    copy_adapter_weights(llama_model, orig_llama_model)

    # make the gate non-zero, otherwise the adapter is disabled and the model
    # identical to regular LLaMA
    enable_gate(llama_model)
    enable_gate(orig_llama_model)

    expected = orig_llama_model(token_sample, 0)
    out = llama_model(token_sample)
    assert torch.allclose(out, expected)


@pytest.mark.skipif(sys.platform in ("win32", "darwin"), reason="torch.compile not supported on this platform")
def test_model_compile(lit_llama):
    llama_config = lit_llama.LLaMAConfig(block_size=8, vocab_size=8, n_layer=2, n_head=2, n_embd=4)
    model = lit_llama.LLaMA(llama_config)
    model.apply(model._init_weights)

    model = torch.compile(model)

    sample = torch.randint(model.config.vocab_size, size=(2, model.config.block_size), dtype=torch.int64)
    for _ in range(3):
        _ = model(sample)



# get news on stocks and run sentiment
# save stock, date, sentiment

# import json
import datetime
import dateparser
import sqlalchemy as db
from db import create_db
from sentiment_utils import get_sentiment, sentiment_analyzer
from utils import DATE_FORMAT
from news_utils import get_google_news
from stocks_lists import stocks_list_all


def get_news_sentiments_store_db(stocks_list, from_date, to_date, num_results=10, verbose=False):
	# setup db:
	engine, connection, stocks, dates, news, sentiments = create_db()
	
	for stock in stocks_list:
		stock_news, _, _ = get_google_news(stock, 
				      num_results=num_results, 
					  time_period=[from_date, to_date]
					)
		print('Articles #:', len(stock_news))
		print(stock_news)

		# # save to json so we do not have to get news again:
		# with open("stock_news.json", "w") as text_file:
		#     jsonString = json.dumps(stock_news)
		#     text_file.write(jsonString)
		#     text_file.close()

		# read json
		# jsonString = open("stock_news.json", "r").read()

		# add articles to database:
		# for article in json.loads(jsonString):
		for article in stock_news:
			# print(article)
			# convert date:
			converted_date = dateparser.parse(article['date'])
			# print(article['date'], converted_date)
			# run sentiment analysis:
			txt_sentiment = get_sentiment(article['text'], sentiment_analyzer, max_length=512)
			# print(txt_sentiment)

			# add to news:
			entry = [{'symbol':article['stock'], 
					'title':article['title'], 
					'date':converted_date, 
					'url':article['link'], 
					'source':article['source'], 
					'text':article['text'], 
					'sentiment':txt_sentiment, 
					'embedding':'[0,0,0]'},
					]
			query = db.insert(news)
			ResultProxy = connection.execute(query,entry)

			#add to sentiments:
			entry = [{'symbol':article['stock'], 
					'date':converted_date, 
					'sentiment':txt_sentiment,} 
					]
			query = db.insert(sentiments)
			ResultProxy = connection.execute(query,entry)


	if verbose:
		# query news:
		print('\n\nNEWS DB:\n')
		query = news.select()
		ResultProxy = connection.execute(query)
		ResultSet = ResultProxy.fetchall()
		print(ResultSet)

		# query sentiments:
		print('\n\SENTIMENTS DB:\n')
		query = sentiments.select()
		ResultProxy = connection.execute(query)
		ResultSet = ResultProxy.fetchall()
		print(ResultSet)

# script to run CRON job - get news for all stocks in our list
if __name__ == "__main__":
	# stocks_list = ['AAPL']
	stocks_list = stocks_list_all # get our list of stocks
	date_today = datetime.date.today().strftime(DATE_FORMAT)
	# run sentimet over last 7days (1 week), max 10 articles per symbol
	get_news_sentiments_store_db(stocks_list, 
				  from_date=(datetime.datetime.now() - datetime.timedelta(days=7)).strftime(DATE_FORMAT), 
				  to_date=date_today,
			      num_results=10,
				  verbose=True,
				)


# train a neural net model on Yahoo Finance data API, Google News
# from: https://huggingface.co/dfurman/falcon-40b-chat-oasst1/blob/main/finetune_falcon40b_oasst1_with_bnb_peft.ipynb
# https://huggingface.co/blog/falcon
# https://colab.research.google.com/drive/1n5U13L0Bzhs32QO_bls5jwuZR62GPSwE?usp=sharing

import os
import sys
import time
from pathlib import Path
import shutil

import torch
from transformers import AutoTokenizer, AutoModelForCausalLM, BitsAndBytesConfig
import transformers
from datasets import load_dataset
from peft import PeftModel, PeftConfig, get_peft_model, LoraConfig, LoraConfig, get_peft_model, prepare_model_for_kbit_training
from tqdm import tqdm
import numpy as np
from argparse import ArgumentParser
from torch.utils.data.dataloader import DataLoader
from sklearn.metrics import mean_squared_error 
import re
from datetime import date, datetime
from dateutil.relativedelta import relativedelta
from atradebot import main, news_utils

os.environ["TOKENIZERS_PARALLELISM"] = "false"

device = "cuda" if torch.cuda.is_available() else "cpu"

OUTFOLDER = 'exp4'
IGNORE_INDEX = -100
MICRO_BATCH_SIZE = 4  # change to 4 for 3090
BATCH_SIZE = 16
GRADIENT_ACCUMULATION_STEPS = BATCH_SIZE // MICRO_BATCH_SIZE

INSTRUCTION_KEY = "### Instruction:"
RESPONSE_KEY = "### Response:"
END_KEY = "### End"


def generate_prompt(data_point, mode='train'):
    """generate prompt for training or eval
    https://github.com/tloen/alpaca-lora

    :param data_point: text data with instruction, input, output
    :type data_point: dict{"instruction": str, "input": str, "output": str}
    :param mode: generate prompt for train or eval, defaults to 'train'
    :type mode: str, optional
    :return: prompt to fed model
    :rtype: str
    """    
    prompt = ""
    if data_point["instruction"]:
        prompt = f"""Below is an instruction that describes a task, paired with an input that provides further context. \
            Write a response that appropriately completes the request.
            ### Instruction: {data_point["instruction"]}
            ### Input: {data_point["input"]}\n ### Response:"""
    else:
        return f"""Below is an instruction that describes a task. Write a response that appropriately completes the request.
            ### Instruction: {data_point["instruction"]}\n ### Response:"""
    if mode == 'train':
            prompt += f""" {data_point["output"]} ### End"""
    return prompt 

def get_str2date(response:str):
    """get date from response

    :param response: string following the fin_train.py generate_prompt format
    :type response: str
    :return: date
    :rtype: datetime.date
    """    
    dx = response.find("### Input:")
    match = re.search(r'\d{4}-\d{2}-\d{2}', response[dx:])
    date = datetime.strptime(match.group(), '%Y-%m-%d').date()
    return date

def get_response(sequence, tokenizer):
    """get response from model output sequence

    :param sequence: output sequence from model
    :type sequence: str
    :param tokenizer: model tokenizer
    :type tokenizer: Tokenizer
    :return: only response part of the sequence
    :rtype: str
    """    
    decoded_str = tokenizer.decode(sequence)
    resp = decoded_str.find(RESPONSE_KEY)
    end = decoded_str[resp:].find(END_KEY)
    if resp == -1 and end != -1:
        return decoded_str[: resp + end]
    elif resp == -1 and end == -1:
        return ''
    else:
        return decoded_str[resp + len(RESPONSE_KEY) : resp + end]

def get_str2alloc(response:str):
    """get allocation for backtesting from response

    :param response: response from model
    :type response: str
    :return: allocation generated from response
    :rtype: dict{stock str: stocks buy/sell int}
    """    
    nums = re.findall(r"[-+]?(?:\d*\.*\d+)", response)
    keys = re.findall(r'\'(.*?)\'', response)
    alloc = {}
    for i, k in enumerate(keys):
        alloc[k] = nums[i]
    return alloc


def get_slm_model(model_name):
    """get a small language model

    :param model_name: id of model from huggingface
    :type model_name: str
    :return:  model, tokenizer
    :rtype:  ModelForCausalLM, Tokenizer
    """    
    model = AutoModelForCausalLM.from_pretrained(model_name)
    tokenizer = AutoTokenizer.from_pretrained("gpt2", padding_side="left")
    tokenizer.pad_token = tokenizer.eos_token
    return model, tokenizer

def get_peft_model(model_name, mode='eval'):
    """get a model with peft 
    :param model_id: id of model from huggingface
    :type model_id: str
    :param mode: eval or train mode, defaults to 'eval'
    :type mode: str, optional
    :return: model, tokenizer
    :rtype: peft ModelForCausalLM, Tokenizer
    """    
    if mode == 'eval':
        config = PeftConfig.from_pretrained(model_name)
        model_id = config.base_model_name_or_path
    else:
        config = LoraConfig(
            r=8, 
            lora_alpha=32, 
            target_modules=["query_key_value"], #gpt_neox
            lora_dropout=0.05, 
            bias="none", 
            task_type="CAUSAL_LM"
        )
    #load model and tokenizer with quantization
    bnb_config = BitsAndBytesConfig(
        load_in_4bit=True,
        bnb_4bit_use_double_quant=True,
        bnb_4bit_quant_type="nf4",
        bnb_4bit_compute_dtype=torch.bfloat16
    )
    tokenizer = AutoTokenizer.from_pretrained(model_id, padding_side="left")
    tokenizer.pad_token = tokenizer.eos_token
    model = AutoModelForCausalLM.from_pretrained(
        model_id, 
        quantization_config=bnb_config, 
        device_map="auto",
        trust_remote_code=True)
    #setup lora
    if mode == 'eval':
        model = PeftModel.from_pretrained(model, model_name)
    else:
        model.gradient_checkpointing_enable()
        model = prepare_model_for_kbit_training(model)
        model = get_peft_model(model, config)

    return model, tokenizer


def train_model(args):
    if args.mode == 'train':
        if args.modeltype == 'small':
            model, tokenizer = get_slm_model("gpt2")
        else:
            model, tokenizer = get_peft_model("databricks/dolly-v2-3b", mode='train')
    else:
        if args.modeltype == 'small':
            model, tokenizer = get_slm_model(args.mhub)
        else:
            model, tokenizer = get_peft_model(args.mhub)
        model.eval()

    #get data
    data = load_dataset(args.dhub)

    data = data.shuffle().map(
        lambda data_point: tokenizer(
            generate_prompt(data_point, mode=args.mode),
            truncation=True,
            max_length=512,
            padding="max_length",
        )
    )
    data = data["train"].train_test_split(test_size=0.1)

    #train: sequence of text and predict next token
    training_args = transformers.TrainingArguments(
        # auto_find_batch_size=True,
        per_device_train_batch_size=MICRO_BATCH_SIZE,
        gradient_accumulation_steps=GRADIENT_ACCUMULATION_STEPS,
        num_train_epochs=100,
        learning_rate=2e-5,
        fp16=True,
        save_total_limit=4,
        report_to='tensorboard',
        logging_dir=os.path.join(OUTFOLDER, "logs"),
        logging_steps=4,
        output_dir=os.path.join(OUTFOLDER, "outputs"),
        save_strategy='epoch',
        # optim="paged_adamw_8bit",
        # lr_scheduler_type = 'cosine',
        # evaluation_strategy = 'epoch',
        warmup_ratio = 0.01,
    )

    trainer = transformers.Trainer(
        model=model,
        train_dataset=data['train'],
        eval_dataset=data['test'],
        args=training_args,
        data_collator=transformers.DataCollatorForLanguageModeling(tokenizer, mlm=False),
    )

    model.config.use_cache = False  # silence the warnings. Please re-enable for inference!
    if args.mode == 'train':
        trainer.train(resume_from_checkpoint = args.reload)
        
        metrics = trainer.evaluate(eval_dataset=data['test'])
        print(metrics)
        #save model
        model.push_to_hub(args.mhub)
        model.save_pretrained(f'{OUTFOLDER}/best_model')

    else:
        model.to(device)
        data.set_format("torch")
        eval_dataloader = DataLoader(data["test"], batch_size=1)

        all_preds, all_targets = [], []
        for in_data in tqdm(eval_dataloader):
            in_data['input_ids'] = in_data['input_ids'].to(device)
            in_data['attention_mask'] = in_data['attention_mask'].to(device)
            with torch.cuda.amp.autocast():
                outputs = model.generate(
                    input_ids = in_data['input_ids'],
                    attention_mask = in_data['attention_mask'], 
                    max_new_tokens=32,
                    pad_token_id=tokenizer.eos_token_id,
                    eos_token_id=tokenizer.eos_token_id)
            for output in outputs:
                response = get_response(output.cpu().numpy(), tokenizer)
                target = in_data['output'][0] 
                print("sample: ", response)
                print("target: ", target)
                all_preds += response
                all_targets += target
        
        # metric_forecast_task(all_targets, all_preds) #for forecast models

def metric_forecast_task(all_targets, all_preds):
    atgt, apred = [], []
    for target, response in zip(all_targets, all_preds):
        pred = re.findall(r"[-+]?(?:\d*\.*\d+)", response)
        tgt = re.findall(r"[-+]?(?:\d*\.*\d+)", target)     
        if len(pred) > 3:
            pred = pred[:3]
        if len(pred) < 3:
            continue
        atgt += tgt
        apred += pred
    atgt = [eval(i) for i in atgt]
    apred = [eval(i) for i in apred]
    print("MSE: ", mean_squared_error(atgt, apred))

def get_parser(raw_args=None):
    parser = ArgumentParser(description="train model")
    parser.add_argument('--reload', action="store_true",
                        help='to reload checkpoint')
    parser.add_argument('--mode', type=str, default='eval',
                        help='train or eval')   
    parser.add_argument('--modeltype', type=str, default='small',
                        help='small or large')   
    parser.add_argument('-d', '--dhub', type=str,
                        default='achang/stocks_one_nvda', help='get from hub folder name for task dataset')
    parser.add_argument('-m', '--mhub', type=str,
                        default='achang/fin_gpt2_one_nvda', help='push to hub folder model')
    args = parser.parse_args(raw_args)
    return args


if __name__ == "__main__":
    args = get_parser()
    train_model(args)



# add different source to gather news and data here
# function input: stock, date, num_results
# function output: list of dict {"link", "title", "snippet", "date", "source", "text", "stock"}

import pandas as pd
# import tweepy
from argparse import ArgumentParser
import os
from datetime import date, datetime, timedelta
import yfinance as yf
import requests
from bs4 import BeautifulSoup
import trafilatura
from trafilatura.settings import use_config
import finnhub

trafilatura_config = use_config()
trafilatura_config.set("DEFAULT", "EXTRACTION_TIMEOUT", "0")
DATE_FORMAT = "%Y-%m-%d"
finnhub_client = finnhub.Client(api_key="cic7bdhr01ql0uqkr15gcic7bdhr01ql0uqkr160")

def get_google_news(stock, num_results=10, time_period=[]):
    """collect google search text 
    limit to 100 results per day

    :param stock: list of stock ids
    :type stock: List[str]
    :param num_results: number of news to collect, defaults to 10
    :type num_results: int, optional
    :param time_period: time_period=['2019-06-28' (start_time), '2019-06-29' (end_time)], defaults to []
    :type time_period: list, optional
    :return: list of news dict {"link", "title", "snippet", "date", "source", "text", "stock"}, search query, soup html searched
    :rtype: list of dict, str, html
    """     
    query = stock
    headers = {
            "User-Agent":
            "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/101.0.4951.54 Safari/537.36"
        }
    if time_period:
        query += f"+before%3A{time_period[1]}+after%3A{time_period[0]}" # add time range     
    search_req = "https://www.google.com/search?q="+query+"&gl=us&tbm=nws&num="+str(num_results)+""
    #https://developers.google.com/custom-search/docs/xml_results#WebSearch_Request_Format
    news_results = []
    # get webpage content
    response = requests.get(search_req, headers=headers)
    soup = BeautifulSoup(response.content, "html.parser")
    for el in soup.select("div.SoaBEf"):
        sublink = el.find("a")["href"]
        downloaded = trafilatura.fetch_url(sublink)
        html_text = trafilatura.extract(downloaded, config=trafilatura_config)
        if html_text:
            news_results.append(
                {
                    "link": el.find("a")["href"],
                    "title": el.select_one("div.MBeuO").get_text(),
                    "snippet": el.select_one(".GI74Re").get_text(),
                    "date": el.select_one(".LfVVr").get_text(),
                    "source": el.select_one(".NUnG9d span").get_text(),
                    "text": html_text,
                    "stock": stock,
                }
            )
    return news_results, search_req, soup

def get_finhub_news(stock, num_results=10, time_period=[]):
    """collect from finhub news: https://finnhub.io/
    only up to 1 year old news and 60 requests per minute

    :param stock: list of stock ids
    :type stock: List[str]
    :param num_results: number of news to collect, defaults to 10
    :type num_results: int, optional
    :param time_period: time_period=['2019-06-28' (start_time), '2019-06-29' (end_time)], defaults to []
    :type time_period: list, optional
    :return: list of news in a dict{"link", "title", "snippet", "date", "source", "text", "stock"}
    :rtype: list of dict, None, None
    """    
    results = finnhub_client.company_news(stock, _from=time_period[0], to=time_period[1])
    news_results = []
    for result in results:
        try:
            downloaded = trafilatura.fetch_url(result['url'])
        except:
            continue
        html_text = trafilatura.extract(downloaded, config=trafilatura_config)
        if html_text:
            news_results.append(
                {
                    "link": result['url'],
                    "title": result['headline'],
                    "snippet": result['summary'],
                    "date": datetime.fromtimestamp(result['datetime']),
                    "source": result['source'],
                    "text": html_text,
                    "stock": stock,
                }
            )
    search_req, soup = None, None
    return news_results, search_req, soup

def get_news(stock, time_period, num_results=10, news_source='google'):
    """wrapper function to get news from different sources

    :param stock: list of stock ids
    :type stock: List[str]
    :param time_period: time_period=['2019-06-28' (start_time), '2019-06-29' (end_time)], defaults to []
    :type time_period: list, optional
    :param num_results: number of news to collect, defaults to 10
    :type num_results: int, optional
    :param news_source: 'google' or 'finhub', defaults to 'finhub'
    :type news_source: str, optional
    :return: list of news dict {"link", "title", "snippet", "date", "source", "text", "stock"}, search query, soup html searched
    :rtype: list of dict, str, html
    """    
    try:
        if news_source == 'google':
            news, _, _ = get_google_news(stock=stock, num_results=num_results, time_period=time_period)
        else:
            news, _, _ = get_finhub_news(stock=stock, num_results=num_results, time_period=time_period)

        if news == []:
            print(f"Can't collect news for {stock} dates {time_period[0]} to {time_period[1]}")
    except:
        print(f"Can't collect news for {stock} dates {time_period[0]} to {time_period[1]}. Error happened.")
        return []

    return news

if __name__ == "__main__":
    news = get_news(stock='AAPL', time_period=['2020-01-01', '2020-01-02'], num_results=3, news_source='google')
    print(news)


import os
import requests
import sqlalchemy as db
import yfinance as yf
import pandas as pd
from sqlalchemy import text
import time
from ib_insync import *

def create_db():
    # create database:
    engine = db.create_engine('sqlite:///atradebot.db', echo=True)
    connection = engine.connect()
    metadata = db.MetaData()

    # create a single table:
    stocks = db.Table('stocks', metadata,
                        db.Column('id', db.Integer(), primary_key=True),
                        db.Column('symbol', db.String(255), nullable=True),
                        db.Column('name', db.String(255), nullable=True),
                        db.Column('sector', db.String(255), nullable=True),
                        db.Column('industry', db.String(255), nullable=True),
                        db.Column('live_price', db.Float(), nullable=True),
                        db.Column('prev_close', db.Float(), nullable=True),
                        db.Column('open', db.Float(), nullable=True),
                        db.Column("volume", db.Integer(), nullable=True),
                    )
    news = db.Table('news', metadata,
                        db.Column('id', db.Integer(), primary_key=True),
                        db.Column('symbol', db.String(255), nullable=True),
                        db.Column('title', db.String(255), nullable=True),
                        db.Column('news_date', db.String(255), nullable=True),
                        db.Column('url', db.String(255), nullable=True),
                        db.Column('source', db.String(255), nullable=True),
                        db.Column('text', db.String(255), nullable=True),
                    )
    etfs = db.Table("etfs", metadata, 
                    db.Column("id", db.Integer(), primary_key=True),
                    db.Column("symbol", db.String(255), nullable=True),
                    db.Column('name', db.String(255), nullable=True),
                    db.Column('sector', db.String(255), nullable=True),
                    db.Column('industry', db.String(255), nullable=True),
                    db.Column('live_price', db.Float(), nullable=True),
                    db.Column('prev_close', db.Float(), nullable=True),
                    db.Column('open', db.Float(), nullable=True),
                    db.Column("volume", db.Integer(), nullable=True),
                    )
                    
    

    # create table in database:
    metadata.create_all(engine)
    return engine, connection, stocks, news

if __name__ == "__main__":
    engine, connection, stocks, news = create_db()

    connection.execute(text("PRAGMA journal_mode=DELETE"))

    # get list of stocks:
    stock_df = pd.read_excel('SP_500_Companies.xlsx')
    symbols = stock_df['Symbol'].tolist()

    ib = IB()
    ib.connect('127.0.0.1', 7496, clientId=1)
    news_providers = ib.reqNewsProviders()
    codes = '+'.join(news_provider.code for news_provider in news_providers)

    # get list of etfs (top 10)
    etfs_symbols = ["SPY", "IVV", "VOO", "VTI", "QQQ", "VEA", "VTV", "IEFA", "BND", "VUG"]

    # get data for symbols:
    for i in symbols[:5]:
        trans = connection.begin_nested()
        try:
            # get stock info
            try:
                ticker = yf.Ticker(i)
                info = ticker.info

                stock_values = {
                    "symbol": i,
                    'name': info.get('shortName', 'NA'),
                    'sector': info.get('sector', 'NA'),
                    'industry': info.get('industry', 'NA'),
                    'live_price': info.get('regularMarketPrice', 0.0),
                    'prev_close': info.get('previousClose', 0.0),
                    'open': info.get('open', 0.0),
                    'volume': info.get('volume', 0)
                }

                # Insert the stock into the database
                query_stock = db.insert(stocks)
                ResultProxy = connection.execute(query_stock, [stock_values])

            # Error handling for stock info fetching
            except requests.exceptions.HTTPError as error_terminal:
                print("HTTPError while fetching stock info for symbol:", i)
            except Exception as e:  # General exception catch
                print("Unexpected error while fetching stock info for symbol:", i)

            # Fetch and store news articles
            try:
                stock = Stock(i, 'SMART', 'USD')
                ib.qualifyContracts(stock)
                headlines = ib.reqHistoricalNews(stock.conId, codes, '', '', 100)

                for headline in headlines:
                    article_date = headline.time.date()
                    article = ib.reqNewsArticle(headline.providerCode, headline.articleId)

                    # Insert the article into the database
                    news_info = {
                        'symbol': i,
                        'title': '',  # Title not needed
                        'news_date': str(article_date),
                        'url': '',  # URL not provided
                        'source': '',  # Source not provided
                        'text': article.articleText
                    }

                    # Insert the news into the database
                    query_news = db.insert(news)
                    ResultProxy = connection.execute(query_news, [news_info])

            # Error handling for news fetching
            except requests.exceptions.HTTPError as error_terminal:
                print("HTTPError while fetching news for symbol:", i)
            except Exception as e:  # General exception catch
                print("Unexpected error while fetching news for symbol:", i)


            trans.commit()
            time.sleep(1)

        # Error handling
        except Exception as e:  # General exception catch
            print("Unexpected error:", e)
            trans.rollback()
    
    # get data for etf symbols
    for i in etfs_symbols:
        trans = connection.begin_nested()
        try: 
            try:
                #get etf info
                ticker = yf.Ticker(i)
                info = ticker.info()

                etf_values = {
                    "symbol": i,
                    'name': info.get('shortName', 'NA'),
                    'sector': info.get('sector', 'NA'),
                    'industry': info.get('industry', 'NA'),
                    'live_price': info.get('regularMarketPrice', 0.0),
                    'prev_close': info.get('previousClose', 0.0),
                    'open': info.get('open', 0.0),
                    'volume': info.get('volume', 0)
                }

                # Insert etfs into the database:
                query_etf = db.insert(etfs)
                ResultProxy = connection.execute(query_etf, [etf_values])

            # Error handling for ETF info fetching:
            except requests.exceptions.HTTPError as error_terminal:
                print("HTTPError while fetching stock info for symbol:", i)
            except Exception as e:  # General exception catch
                print("Unexpected error while fetching stock info for symbol:", i)

            trans.commit()
            time.sleep(1)
        # Error handling:
        except Exception as e:  # General exception catch
            print("Unexpected error:", e)
            trans.rollback()


    # Fetch and print the first 5 stocks from the database after processing
    #query = db.select([data]).limit(5)
    query = stocks.select().limit(5)
    ResultProxy = connection.execute(query)
    ResultSet = ResultProxy.fetchall()

    for result in ResultSet:
        print(result)

    connection.execute(text("PRAGMA journal_mode=WAL"))

    # Close the connection
    connection.close()



# cron job for getting news to db

import json
import sqlalchemy as db
from atradebot.db import create_db
from atradebot.news_utils import get_google_news

# setup db:
engine, connection, stocks, dates, news, sentiments = create_db()

# get some news for 1 stock:
stock = 'AAPL'
stock_news, _, _ = get_google_news(stock, num_results=10, time_period=['2022-06-28', '2022-06-31'])
print('Articles #:', len(stock_news))
print(stock_news)

# # save to json so we do not have to get news again:
# with open("stock_news.json", "w") as text_file:
#     jsonString = json.dumps(stock_news)
#     text_file.write(jsonString)
#     text_file.close()

# read json
# jsonString = open("stock_news.json", "r").read()

# add articles to database:
# for article in json.loads(jsonString):
for article in stock_news:
    print(article)
    entry = [{'symbol':article['stock'], 
              'title':article['title'], 
              'date':article['date'], 
              'url':article['link'], 
              'source':article['source'], 
              'text':article['text'], 
              'sentiment':'0', 
              'embedding':'[0,0,0]'},
            ]
    query = db.insert(news)
    ResultProxy = connection.execute(query,entry)

# query news:
query = news.select()
ResultProxy = connection.execute(query)
ResultSet = ResultProxy.fetchall()
print(ResultSet)


import tempfile
import pathlib

import torch


class ATensor(torch.Tensor):
    pass


def test_lazy_load_basic(lit_llama):
    import lit_llama.utils

    with tempfile.TemporaryDirectory() as tmpdirname:
        m = torch.nn.Linear(5, 3)
        path = pathlib.Path(tmpdirname)
        fn = str(path / "test.pt")
        torch.save(m.state_dict(), fn)
        with lit_llama.utils.lazy_load(fn) as sd_lazy:
            assert "NotYetLoadedTensor" in str(next(iter(sd_lazy.values())))
            m2 = torch.nn.Linear(5, 3)
            m2.load_state_dict(sd_lazy)

        x = torch.randn(2, 5)
        actual = m2(x)
        expected = m(x)
        torch.testing.assert_close(actual, expected)


def test_lazy_load_subclass(lit_llama):
    import lit_llama.utils

    with tempfile.TemporaryDirectory() as tmpdirname:
        path = pathlib.Path(tmpdirname)
        fn = str(path / "test.pt")
        t = torch.randn(2, 3)[:, 1:]
        sd = {
            1: t,
            2: torch.nn.Parameter(t),
            3: torch.Tensor._make_subclass(ATensor, t),
        }
        torch.save(sd, fn)
        with lit_llama.utils.lazy_load(fn) as sd_lazy:
            for k in sd.keys():
                actual = sd_lazy[k]
                expected = sd[k]
                torch.testing.assert_close(actual._load_tensor(), expected)


def test_incremental_write(tmp_path, lit_llama):
    import lit_llama.utils

    sd = {str(k): torch.randn(5, 10) for k in range(3)}
    sd_expected = {k: v.clone() for k, v in sd.items()}
    fn = str(tmp_path / "test.pt")
    with lit_llama.utils.incremental_save(fn) as f:
        sd["0"] = f.store_early(sd["0"])
        sd["2"] = f.store_early(sd["2"])
        f.save(sd)
    sd_actual = torch.load(fn)
    assert sd_actual.keys() == sd_expected.keys()
    for k, v_expected in sd_expected.items():
        v_actual = sd_actual[k]
        torch.testing.assert_close(v_expected, v_actual)


def test_find_multiple(lit_llama):
    from lit_llama.utils import find_multiple

    assert find_multiple(17, 5) == 20
    assert find_multiple(30, 7) == 35
    assert find_multiple(10, 2) == 10
    assert find_multiple(5, 10) == 10



"""Implementation derived from https://github.com/tloen/alpaca-lora"""
import sys
from pathlib import Path

# support running without installing as a package
wd = Path(__file__).parent.parent.resolve()
sys.path.append(str(wd))

import torch
import requests
import json
from torch.utils.data import random_split
from lit_llama.tokenizer import Tokenizer
from tqdm import tqdm


IGNORE_INDEX = -1

DATA_FILE_NAME = "input.txt"


def prepare(
    destination_path: Path = Path("data/any"),
    tokenizer_path: Path = Path("checkpoints/lit-llama/tokenizer.model"),
    test_split_ratio: float = 0.9,  # default 90% train, 10% validation
    max_seq_length: int = 256,
    seed: int = 42,
    data_file_name: str = DATA_FILE_NAME,
) -> None:
    """Prepare any dataset for finetuning (akin to Shakespheare full tuning).

    The output is a training and validation dataset saved as `train.pt` and `val.pt`,
    which stores the preprocessed and tokenized prompts and labels.
    """

    destination_path.mkdir(parents=True, exist_ok=True)
    file_path = destination_path / data_file_name
    if not file_path.exists():
        raise AssertionError(f"{data_file_name} is provided by the user")

    # TODO: If we don't have the Meta weights, where do we get the tokenizer from?
    tokenizer = Tokenizer(tokenizer_path)

    data = []

    with open(file_path, "r") as input_file:
        for line in input_file.readlines():
            data.append(line)

    # Partition the dataset into train and test
    train_split_size = int(len(data) * test_split_ratio)
    test_split_size = len(data) - train_split_size
    train_set, test_set = random_split(
        data,
        lengths=(train_split_size, test_split_size),
        generator=torch.Generator().manual_seed(seed),
    )
    train_set, test_set = list(train_set), list(test_set)

    print(f"train has {len(train_set):,} samples")
    print(f"val has {len(test_set):,} samples")

    print("Processing train split ...")
    train_set = [
        prepare_line(line, tokenizer, max_seq_length) for line in tqdm(train_set)
    ]
    torch.save(train_set, file_path.parent / "train.pt")

    print("Processing test split ...")
    test_set = [
        prepare_line(line, tokenizer, max_seq_length) for line in tqdm(test_set)
    ]
    torch.save(test_set, file_path.parent / "test.pt")


def prepare_line(line: str, tokenizer: Tokenizer, max_length: int):
    """Processes a single sample.

    This function processes the line to produce the tokenized version of it.
    """
    encoded_full_prompt = tokenize(tokenizer, line, max_length=max_length, eos=False)
    return {
        "input_ids": encoded_full_prompt,
        "labels": encoded_full_prompt,
    }


def tokenize(
    tokenizer: Tokenizer, string: str, max_length: int, eos=True
) -> torch.Tensor:
    return tokenizer.encode(string, bos=True, eos=eos, max_length=max_length)


if __name__ == "__main__":
    from jsonargparse import CLI

    CLI(prepare)



import sys
from pathlib import Path

import pytest

wd = Path(__file__).parent.parent.absolute()


@pytest.fixture()
def orig_llama():
    sys.path.append(str(wd))

    from scripts.download import download_original

    download_original(wd)

    import original_model

    return original_model


@pytest.fixture()
def orig_llama_adapter():
    sys.path.append(str(wd))

    from scripts.download import download_original

    download_original(wd)

    import original_adapter

    return original_adapter


@pytest.fixture()
def lit_llama():
    # this adds support for running tests without the package installed
    sys.path.append(str(wd))

    import lit_llama

    return lit_llama
